 /***********************************************************
  *   Program:    LCD driver
  *   Created:    28.8.99 21:32
  *   Author:     Brajer Vlado
  *                 vlado.brajer@kks.s-net.net
  *   Comments:   HITACHI LCD driver
  *
  *   Remarks:    20020513 Stephan Linz <linz@li-pro.net>
  *               adapted to AVR-Ctrl board designed by
  *               http://www.mikrocontroller.com
  *
  ************************************************************/
#include <libac.h>

#define LCD_CMD_DISPLAY_CLEAR		(BV(0))
#define LCD_CMD_CURSOR_HOME		(BV(1))
//#define LCD_CMD_ENTRY	... TODO
#define LCD_CMD_DISPLAY_ON		(BV(3) | BV(2))
#define LCD_CMD_DISPLAY_OFF		(BV(3) & ~BV(2))
#define LCD_CMD_CURSOR_ON		(BV(3) | BV(1))
#define LCD_CMD_CURSOR_OFF		(BV(3) & ~BV(1))
#define LCD_CMD_BLINKING_ON		(BV(3) | BV(0))
#define LCD_CMD_BLINKING_OFF		(BV(3) & ~BV(0))
#define LCD_CMD_CGRADDR			(BV(6))
#define LCD_CMD_DDRADDR			(BV(7))

/*
 *
 * DD RAM address values for 2x16 sample (ADD):
 *
 *       0  1  2  3  4  5  6  7  8  9  1  1  1  1  1  1|
 *                                     0  1  2  3  4  5|
 *                                                     |
 *  0-  00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F|10 11 12 13 14 15 16...
 *  1-  40 41 42 43 44 45 46 47 48 49 4A 4B 4C 4D 4E 4F|50 51 52 53 54 55 56...
 *       ^
 *       |
 *      _lcd_base_y[(0..3) & _lcd_mask_y] with _lcd_mask_y = 1
 *
 * DD RAM address values for 4x16 sample (ADD):
 *
 *       0  1  2  3  4  5  6  7  8  9  1  1  1  1  1  1|
 *                                     0  1  2  3  4  5|
 *                                                     |
 *  0-  00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F|
 *  1-  40 41 42 43 44 45 46 47 48 49 4A 4B 4C 4D 4E 4F|
 *  2-  10 11 12 13 14 15 16 17 18 19 1A 1B 1C 1D 1E 1F|20 21 22 23 24 25 26 27
 *  3-  50 51 52 53 54 55 56 57 58 59 5A 5B 5C 5D 5E 5F|50 51 52 53 54 55 56 57
 *       ^
 *       |
 *      _lcd_base_y[(0..3) & _lcd_mask_y] with _lcd_mask_y = 3
 *
 */
static uint8_t _lcd_base_y[4] = { 0x00, 0x40 };
static uint8_t _lcd_mask_y, _lcd_x, _lcd_y, _lcd_maxx;


static void _lcd_out_p(uint8_t b, uint8_t p)
{
  /* leaf non used signal unchanged (LCD_UNUSED_BIT) */
  outp((inp(p) & BV(LCD_UNUSED_BIT)) | (b & ~BV(LCD_UNUSED_BIT)), p);
}

static void _lcd_toggle_EN(void)
{
  sbi((LCD_PORT), LCD_EN);	// EN = 1
  cbi((LCD_PORT), LCD_EN);	// EN = 0
}

#ifdef LCD_COM_POLLING
static void _lcd_check_BF(void)
{
  _lcd_out_p(0x0f, (LCD_PORT) - 1);	// upper nibble as input (temporary)
  sbi((LCD_PORT), LCD_RD);		// RD = 1
  cbi((LCD_PORT), LCD_RS);		// RS = 0
  do
  {
    __delay_6T();
    sbi((LCD_PORT), LCD_EN);		// EN = 1
    __delay_6T();
  } while (bit_is_set((LCD_PORT) - 2, 7));
  cbi((LCD_PORT), LCD_EN);		// EN = 0
  _lcd_out_p(0xff, (LCD_PORT) - 1);	// back to output port (default mode)
}

void _lcd_ready(void)
{
  _lcd_check_BF();
}
#else
# warning "LCD_COM_POLLING not defined -- use simulated _lcd_ready()"
void _lcd_ready(void)
{
  delay_ms(2);
}
#endif

static void _lcd_write_CMD(uint8_t byte)
{
  /* upper nibble */
  _lcd_out_p((byte & 0xf0) & ~BV(LCD_RS) & ~BV(LCD_RD), (LCD_PORT));
  _lcd_toggle_EN();
  /* lower nibble */
  _lcd_out_p((byte << 4) & ~BV(LCD_RS) & ~BV(LCD_RD), (LCD_PORT));
  _lcd_toggle_EN();
}

static void _lcd_write_DAT(uint8_t byte)
{
  /* upper nibble */
  _lcd_out_p(((byte & 0xf0) | BV(LCD_RS)) & ~BV(LCD_RD), (LCD_PORT));
  _lcd_toggle_EN();
  /* lower nibble */
  _lcd_out_p(((byte << 4) | BV(LCD_RS)) & ~BV(LCD_RD), (LCD_PORT));
  _lcd_toggle_EN();
}

void _lcd_write_data(uint8_t data)
{
  _lcd_write_CMD(data);
}

#ifdef LCD_COM_POLLING
void lcd_cls(void)
{
  _lcd_check_BF();
  _lcd_write_CMD(LCD_CMD_DISPLAY_CLEAR);
}

void lcd_home(void)
{
  _lcd_check_BF();
  _lcd_write_CMD(LCD_CMD_CURSOR_HOME);
}

void lcd_goto(uint8_t addr)
{
  _lcd_check_BF();
  _lcd_write_CMD(LCD_CMD_DDRADDR | (addr & 0x7f));
}

void lcd_write_byte(uint8_t addr, uint8_t data)
{
  _lcd_check_BF();
  _lcd_write_CMD(addr);
  _lcd_check_BF();
  _lcd_write_DAT(data);
}
#else
void lcd_cls(void)
{
  _lcd_write_CMD(LCD_CMD_DISPLAY_CLEAR);
  delay_ms(2);
}

void lcd_home(void)
{
  _lcd_write_CMD(LCD_CMD_CURSOR_HOME);
  delay_ms(2);
}

void lcd_goto(uint8_t addr)
{
  _lcd_write_CMD(LCD_CMD_DDRADDR | (addr & 0x7f));
  delay_ms(2);
}

void lcd_write_byte(uint8_t addr, uint8_t data)
{
  _lcd_write_CMD(addr);
  delay_us(50);
  _lcd_write_DAT(data);
  delay_us(50);
}
#endif

/*-----------------27.5.99 20:33--------------------
 *  LCD & Cursor control (Blinking, ON/OFF,...)
 *--------------------------------------------------*/
void lcd_control(uint8_t disonoff, uint8_t curonoff, uint8_t curblink)
{ uint8_t temp = 0;
  if (disonoff==1) temp |= LCD_CMD_DISPLAY_ON; else temp |= LCD_CMD_DISPLAY_OFF;
  if (curonoff==1) temp |= LCD_CMD_CURSOR_ON; else temp |= LCD_CMD_CURSOR_OFF;
  if (curblink==1) temp |= LCD_CMD_BLINKING_ON; else temp |= LCD_CMD_BLINKING_OFF;
#ifdef LCD_COM_POLLING
  _lcd_check_BF();
#endif
  _lcd_write_CMD(temp);
#ifdef LCD_COM_DELAY
  delay_us(50);
#endif
}

void lcd_gotoxy(uint8_t x, uint8_t y)
{
  lcd_goto(_lcd_base_y[(_lcd_y = y & _lcd_mask_y)] + (_lcd_x = x));
}


/*-----------------27.5.99 20:30--------------------
 *  Display initialization
 *--------------------------------------------------*/
void lcd_init(uint8_t lcd_matrix)
{
  _lcd_maxx		= lcd_matrix & 0x3f;	// save for further usage
  _lcd_base_y[2]	= _lcd_maxx;
  _lcd_base_y[3]	= _lcd_maxx + 0x40;

  if ((lcd_matrix >> 6) == 0 || (lcd_matrix >> 6) == 3)
  {
    _lcd_mask_y = 3;
  }
  else
  {
    _lcd_mask_y = lcd_matrix >> 7;
  }

  delay_ms(20);				// avoid power-on mismatches at LCD device
  _lcd_out_p(0xff, (LCD_PORT) - 1);	// output port (default mode)

  _lcd_out_p(0x30, (LCD_PORT)); _lcd_toggle_EN(); delay_ms(5);	// init step 1
  _lcd_out_p(0x30, (LCD_PORT)); _lcd_toggle_EN(); delay_us(110);// init step 2
  _lcd_out_p(0x30, (LCD_PORT)); _lcd_toggle_EN(); delay_us(50);	// init step 3
  _lcd_out_p(0x20, (LCD_PORT)); _lcd_toggle_EN(); delay_us(50);	// init 4bit mode
  _lcd_out_p(0x20, (LCD_PORT)); _lcd_toggle_EN();		// functionset
  _lcd_out_p(0x80, (LCD_PORT)); _lcd_toggle_EN(); delay_us(50);	// 4bit, dual line
  _lcd_out_p(0x00, (LCD_PORT)); _lcd_toggle_EN();		// entry mode:
  _lcd_out_p(0x60, (LCD_PORT)); _lcd_toggle_EN(); delay_us(50);	// increment cursor

  lcd_cls();		// be shure, there is nothing at display :-)
  lcd_home();
  lcd_control(1, 1, 1);	// display, cursor, and blink on
}

/*-----------------27.5.99 20:31--------------------
 *  Put character on LCD
 *--------------------------------------------------*/
void lcd_putch(uint8_t data)
{
#ifdef LCD_COM_POLLING
  _lcd_check_BF();
#endif
  _lcd_write_DAT(data);
#ifdef LCD_COM_DELAY
  delay_us(50);
#endif
  if (++_lcd_x == _lcd_maxx) lcd_gotoxy(0, _lcd_y + 1);
}
